@@ -8,7 +8,7 @@ import pickle
 import importlib
 import copy
 import warnings
-import pytest
+import pytest # type:ignore
 from bs4 import BeautifulSoup
 from bs4.element import (
     AttributeValueList,
@@ -49,10 +49,10 @@ try:
 except ImportError:
     SOUP_SIEVE_PRESENT = False
 
-HTML5LIB_PRESENT = importlib.util.find_spec("html5lib") is not None
+HTML5LIB_PRESENT = importlib.util.find_spec("html5lib") is not None # type:ignore
 
 try:
-    import lxml.etree
+    import lxml.etree # type:ignore
     LXML_PRESENT = True
     LXML_VERSION = lxml.etree.LXML_VERSION
 except ImportError:
@@ -431,11 +431,24 @@ class HTMLTreeBuilderSmokeTest(TreeBuilderSmokeTest):
         # Pickling a tree, then unpickling it, yields a tree identical
         # to the original.
         tree = self.soup("<a><b>foo</a>")
-        dumped = pickle.dumps(tree, 2)
+        dumped = pickle.dumps(tree, pickle.HIGHEST_PROTOCOL)
         loaded = pickle.loads(dumped)
         assert loaded.__class__ == BeautifulSoup
         assert loaded.decode() == tree.decode()
 
+    def test_pickle_and_unpickle_bad_markup(self):
+        markup = """
+<!DOCTYPE html>
+<html lang="en">
+<head><title>blabla</title></head>
+<body><?xml encoding="utf-8" ?><html></html></body>
+</html>
+"""
+        soup = self.soup(markup)
+        pickled = pickle.dumps(soup, pickle.HIGHEST_PROTOCOL)
+        soup = pickle.loads(pickled)
+        assert soup.builder.is_xml is False
+
     def assertDoctypeHandled(self, doctype_fragment: str) -> None:
         """Assert that a given doctype string is handled correctly."""
         doctype_str, soup = self._document_with_doctype(doctype_fragment)
@@ -511,9 +524,17 @@ class HTMLTreeBuilderSmokeTest(TreeBuilderSmokeTest):
 <head><title>Hello.</title></head>
 <body>Goodbye.</body>
 </html>"""
+
+        # libxml 2.14.3 turns the XML declaration into a comment, turning the document
+        # from XHTML to regular HTML. That's fine with this test.
+        markup_as_html = markup.replace(b'<?xml version="1.0" encoding="utf-8"?>',
+                                        b'<!--?xml version="1.0" encoding="utf-8"?-->')
+
         with warnings.catch_warnings(record=True) as w:
             soup = self.soup(markup)
-        assert soup.encode("utf-8").replace(b"\n", b"") == markup.replace(b"\n", b"")
+        without_minor_differences = soup.encode("utf-8").replace(b"\n", b"")
+        assert (without_minor_differences == markup.replace(b"\n", b"")
+                or without_minor_differences == markup_as_html.replace(b"\n", b""))
 
         # No warning was issued about parsing an XML document as HTML,
         # because XHTML is both.
@@ -553,13 +574,18 @@ class HTMLTreeBuilderSmokeTest(TreeBuilderSmokeTest):
         # process_markup correctly sets processing_instruction_class
         # even when the markup is already Unicode and there is no
         # need to process anything.
+        #
+        # Since HTML doesn't have any processing instructions, it's also
+        # legal to convert the PI into a comment, as libxml2 2.14.3 does.
         markup = """<?PITarget PIContent?>"""
+        markup_as_comment = "<!--?PITarget PIContent?-->"
         soup = self.soup(markup)
-        assert markup == soup.decode()
+        assert soup.decode() in [markup, markup_as_comment]
 
-        markup = b"""<?PITarget PIContent?>"""
+        markup = markup.encode("utf8")
+        markup_as_comment = markup_as_comment.encode("utf8")
         soup = self.soup(markup)
-        assert markup == soup.encode("utf8")
+        assert soup.encode("utf8") in [markup, markup_as_comment]
 
     def test_deepcopy(self):
         """Make sure you can copy the tree builder.
@@ -1085,6 +1111,14 @@ Hello, world!
         soup = self.soup("<body><div><p>text1</p></span>text2</div></body>")
         assert "<body><div><p>text1</p>text2</div></body>" == soup.body.decode()
 
+    def test_not_really_a_namespace(self):
+        # There's some markup here that looks a bit like a namespace ({value: 20000.00}) but isn't.
+        markup = '<!DOCTYPE html><html lang="id"><head><noscript><img height="1" width="1" style="display:none" src="http://example.com/" fbq(\'track\', \'Purchase\' , {value: 20000.00}); /></noscript></head><body></body>'
+        soup = self.soup(markup)
+
+        # We don't care what the parser makes of this, but it needs to be able to parse this markup without crashing.
+        soup.body.decode()
+
     def test_worst_case(self):
         """Test the worst case (currently) for linking issues."""
 
